@@ -1,5 +1,8 @@
 init_authcode <- function()
 {
+    stopifnot(is.list(self$token_args))
+    stopifnot(is.list(self$authorize_args))
+
     if(!requireNamespace("httpuv", quietly=TRUE))
         stop("httpuv package must be installed to use authorization_code method", call.=FALSE)
 
@@ -14,7 +17,7 @@ init_authcode <- function()
         scope=paste(self$scope, collapse=" "),
         client_secret=self$client$client_secret,
         login_hint=self$client$login_hint,
-        state=paste0(sample(letters, 20), collapse="") # random nonce
+        state=paste0(sample(letters, 20, TRUE), collapse="") # random nonce
     ), self$authorize_args)
 
     auth_uri$query <- opts
@@ -24,7 +27,7 @@ init_authcode <- function()
 
     # contact token endpoint for token
     access_uri <- private$aad_endpoint("token")
-    body <- c(self$client, code=code, redirect_uri=opts$redirect_uri)
+    body <- c(self$client, code=code, redirect_uri=opts$redirect_uri, self$token_args)
 
     httr::POST(access_uri, body=body, encode="form")
 }
@@ -101,7 +104,10 @@ listen_for_authcode <- function(url, localhost="127.0.0.1", localport=1410)
     httpuv::service()
 
     if(is_empty(info$code))
-        stop("Authentication failed.", call.=FALSE)
+    {
+        msg <- gsub("\\+", " ", utils::URLdecode(info$error_description))
+        stop("Authentication failed. Message:\n", msg, call.=FALSE)
+    }
 
     message("Authentication complete.")
     info$code
